"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from babel.numbers import get_currency_symbol

from biweeklybudget.flaskapp.app import app
from biweeklybudget.flaskapp.notifications import NotificationsController
from biweeklybudget import settings as settingsmod


@app.context_processor
def notifications():
    """
    Add notifications to template context for all templates.

    :return: template context with notifications added
    :rtype: dict
    """
    return dict(notifications=NotificationsController().get_notifications())


@app.context_processor
def settings():
    """
    Add settings to template context for all templates.

    :return: template context with settings added
    :rtype: dict
    """
    return {'settings': {x: getattr(settingsmod, x) for x in dir(settingsmod)}}


@app.context_processor
def utilities():
    """
    Utility functions to put in the jinja context.

    :return: template context with utility functions added
    :rtype: dict
    """
    def cast_float(x):
        return float(x)
    return dict(cast_float=cast_float)


@app.context_processor
def add_currency_symbol():
    """
    Context processor to inject the proper currency symbol into the Jinja2
    context as the "CURRENCY_SYM" variable.

    :return: proper currency symbol for our locale and currency
    :rtype: str
    """
    return dict(CURRENCY_SYM=get_currency_symbol(
        settingsmod.CURRENCY_CODE, locale=settingsmod.LOCALE_NAME
    ))



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging

from flask import Flask

from biweeklybudget.db import init_db, cleanup_db
from biweeklybudget.flaskapp.jsonencoder import MagicJSONEncoder
from biweeklybudget.utils import fix_werkzeug_logger

format = "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - " \
         "%(name)s.%(funcName)s() ] %(message)s"
logging.basicConfig(level=logging.DEBUG, format=format)
logger = logging.getLogger()

fix_werkzeug_logger()

app = Flask(__name__)
app.jinja_env.add_extension('jinja2.ext.loopcontrols')
app.json_encoder = MagicJSONEncoder
init_db()


def before_request():
    """
    When running in debug mode, clear jinja cache.
    """
    logger.warning('DEBUG MODE - Clearing jinja cache')
    app.jinja_env.cache = {}


@app.teardown_appcontext
def shutdown_session(exception=None):
    cleanup_db()


if app.debug:
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.before_request(before_request)
    app.jinja_env.auto_reload = True


from biweeklybudget.flaskapp.views import *  # noqa
from biweeklybudget.flaskapp.filters import *  # noqa
from biweeklybudget.flaskapp.jinja_tests import *  # noqa
from biweeklybudget.flaskapp.context_processors import *  # noqa
from biweeklybudget.flaskapp.cli_commands import *  # noqa



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
import json
from decimal import Decimal, ROUND_UP
from datetime import datetime

from flask.views import MethodView
from flask import render_template, request, jsonify

from biweeklybudget.flaskapp.jsonencoder import MagicJSONEncoder
from biweeklybudget.flaskapp.app import app
from biweeklybudget.db import db_session
from biweeklybudget.interest import InterestHelper
from biweeklybudget.models.dbsetting import DBSetting
from biweeklybudget.utils import fmt_currency

logger = logging.getLogger(__name__)


def parse_payoff_settings_json(j):
    """
    Given the 'credit-payoff' DBSettings JSON string value, parse it and return
    a dict with keys "increases" and "onetimes", each having a value of a dict
    of :py:class:`datetime.date` to :py:class:`decimal.Decimal`.

    :param j: 'credit-payoff' DBSettings JSON string value
    :type j: str
    :return: credit payoff settings
    :rtype: dict
    """
    pass


class CreditPayoffsView(MethodView):
    """
    Render the top-level GET /accounts/credit-payoff view using
    ``credit-payoffs.html`` template.
    """

    def _payoffs_list(self, ih):
        """
        Return a payoffs list suitable for rendering.

        :param ih: interest helper instance
        :type ih: biweeklybudget.interest.InterestHelper
        :return: list of payoffs suitable for rendering
        :rtype: list
        """
        res = ih.calculate_payoffs()
        payoffs = []
        for methname in sorted(res.keys(), reverse=True):
            tmp = {
                'name': methname,
                'description': res[methname]['description'],
                'doc': res[methname]['doc'],
                'results': []
            }
            if 'error' in res[methname]:
                tmp['error'] = res[methname]['error']
                continue
            total_pymt = Decimal('0')
            total_int = Decimal('0')
            max_mos = 0
            for k in sorted(res[methname]['results'].keys()):
                r = res[methname]['results'][k]
                acct = ih.accounts[k]
                tmp['results'].append({
                    'name': '%s (%d) (%s @ %s%%)' % (
                        acct.name, k,
                        fmt_currency(abs(acct.balance.ledger)),
                        (acct.effective_apr * Decimal('100')).quantize(
                            Decimal('.01')
                        )
                    ),
                    'total_payments': r['total_payments'],
                    'total_interest': r['total_interest'],
                    'payoff_months': r['payoff_months']
                })
                total_pymt += r['total_payments']
                total_int += r['total_interest']
                if r['payoff_months'] > max_mos:
                    max_mos = r['payoff_months']
            tmp['total'] = {
                'total_payments': total_pymt,
                'total_interest': total_int,
                'payoff_months': max_mos
            }
            payoffs.append(tmp)
        return payoffs

    def _payment_settings_dict(self, settings_json):
        """
        Given the JSON string payment settings, return a dict of payment
        settings as expected by :py:class:`~.InterestHelper` kwargs.

        :param settings_json: payment settings JSON
        :type settings_json: str
        :return: payment settings dict
        :rtype: dict
        """
        res = {'increases': {}, 'onetimes': {}}
        j = json.loads(settings_json)
        for i in j['increases']:
            if not i['enabled']:
                continue
            d = datetime.strptime(i['date'], '%Y-%m-%d').date()
            res['increases'][d] = Decimal(i['amount'])
        for i in j['onetimes']:
            if not i['enabled']:
                continue
            d = datetime.strptime(i['date'], '%Y-%m-%d').date()
            res['onetimes'][d] = Decimal(i['amount'])
        return res

    def get(self):
        setting = db_session.query(DBSetting).get('credit-payoff')
        if setting is None:
            pymt_settings_json = json.dumps({'increases': [], 'onetimes': []})
        else:
            pymt_settings_json = setting.value
        pymt_settings_kwargs = self._payment_settings_dict(pymt_settings_json)
        ih = InterestHelper(db_session, **pymt_settings_kwargs)
        mps = sum(ih.min_payments.values())
        payoffs = self._payoffs_list(ih)
        return render_template(
            'credit-payoffs.html',
            monthly_pymt_sum=mps.quantize(Decimal('.01'), rounding=ROUND_UP),
            payoffs=payoffs,
            pymt_settings_json=pymt_settings_json
        )


class PayoffSettingsFormHandler(MethodView):
    """
    Handle POST /settings/credit-payoff
    """

    def post(self):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        data = request.get_json(force=True, silent=True)
        if data is None:
            logger.error('Error parsing request JSON')
            return jsonify({
                'success': False,
                'error_message': 'Error parsing JSON'
            })
        setting = db_session.query(DBSetting).get('credit-payoff')
        if setting is None:
            setting = DBSetting(name='credit-payoff')
            logger.info('new DBSetting name=credit-payoff')
        else:
            logger.info('Existing DBSetting name=credit-payoff value=%s',
                        setting.value)
        fixeddata = {'increases': [], 'onetimes': []}
        for key in ['increases', 'onetimes']:
            for d in sorted(data[key], key=lambda k: k['date']):
                if d['date'] == '' or d['amount'] == '':
                    continue
                fixeddata[key].append(d)
        val = json.dumps(fixeddata, sort_keys=True, cls=MagicJSONEncoder)
        try:
            parse_payoff_settings_json(val)
        except Exception as ex:
            logger.error('Error converting payoff settings JSON', exc_info=True)
            return jsonify({
                'success': False,
                'error_message': 'Error parsing JSON: %s' % ex
            })
        logger.info('Changing setting value to: %s', val)
        setting.value = val
        db_session.add(setting)
        db_session.commit()
        return jsonify({
            'success': True,
            'success_message': 'Successfully updated setting '
                               '"credit-payoff" in database.'
        })


app.add_url_rule(
    '/accounts/credit-payoff',
    view_func=CreditPayoffsView.as_view('credit_payoffs_view')
)
app.add_url_rule(
    '/settings/credit-payoff',
    view_func=PayoffSettingsFormHandler.as_view('payoff_settings_form')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from decimal import Decimal
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.account import Account, AcctType
from selenium.webdriver.support.ui import Select


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestAccountsNavigation(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts')

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'Accounts - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestAccountsMainPage(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/accounts')

    def test_bank_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-bank-accounts']//table"
        )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Unreconciled', 'Difference'
        ]
        assert self.tbody2textlist(table) == [
            ['BankOne', '$12,789.01 (14 hours ago)', '$0.00', '$12,789.01'],
            ['BankTwoStale', '$100.23 (18 days ago)', '-$333.33', '$433.56']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="javascript:accountModal(1, null)">BankOne</a>',
            '<a href="javascript:accountModal(2, null)">BankTwoStale</a>',
        ]

    def test_bank_stale_span(self, selenium):
        tbody = selenium.find_element_by_xpath(
            "//div[@id='panel-bank-accounts']//table/tbody"
        )
        rows = tbody.find_elements_by_tag_name('tr')
        bankTwoStale_bal_td = rows[1].find_elements_by_tag_name('td')[1]
        bal_span = bankTwoStale_bal_td.find_elements_by_tag_name('span')[1]
        assert bal_span.text == '(18 days ago)'
        assert bal_span.get_attribute('class') == 'data_age text-danger'

    def test_credit_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-credit-cards']//table"
        )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Credit Limit', 'Available', 'Unreconciled',
            'Difference'
        ]
        assert self.tbody2textlist(table) == [
            [
                'CreditOne', '-$952.06 (13 hours ago)', '$2,000.00',
                '$1,047.94', '$222.22', '$825.72'
            ],
            [
                'CreditTwo', '-$5,498.65 (a day ago)', '$5,500.00', '$1.35',
                '$0.00', '$1.35'
            ]
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="javascript:accountModal(3, null)">CreditOne</a>',
            '<a href="javascript:accountModal(4, null)">CreditTwo</a>',
        ]

    def test_investment_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-investment']//table"
        )
        assert self.thead2list(table) == ['Account', 'Value']
        assert self.tbody2textlist(table) == [
            ['InvestmentOne', '$10,362.91 (13 days ago)']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="javascript:accountModal(5, null)">InvestmentOne</a>'
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestAccountModal(AcceptanceHelper):

    def test_10_verify_db(self, testdb):
        acct = testdb.query(Account).get(1)
        assert acct is not None
        assert acct.name == 'BankOne'
        assert acct.description == 'First Bank Account'
        assert acct.ofx_cat_memo_to_name is True
        assert acct.vault_creds_path == 'secret/foo/bar/BankOne'
        assert acct.ofxgetter_config_json == '{"foo": "bar"}'
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Bank
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is True

    def test_11_get_acct1_url(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/1')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 1'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == '1'
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == 'BankOne'
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == 'First Bank Account'
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected()
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected()
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == 'secret/foo/bar/BankOne'
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == '{"foo": "bar"}'
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected() is True
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed() is False
        # END CREDIT
        assert selenium.find_element_by_id('account_frm_active').is_selected()

    def test_12_edit_acct1(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/1')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 1'
        selenium.find_element_by_id('account_frm_name').send_keys('Edited')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 1 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_13_verify_db(self, testdb):
        acct = testdb.query(Account).get(1)
        assert acct is not None
        assert acct.name == 'BankOneEdited'
        assert acct.description == 'First Bank Account'
        assert acct.ofx_cat_memo_to_name is True
        assert acct.vault_creds_path == 'secret/foo/bar/BankOne'
        assert acct.ofxgetter_config_json == '{"foo": "bar"}'
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Bank
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is True

    def test_20_verify_db(self, testdb):
        acct = testdb.query(Account).get(2)
        assert acct is not None
        assert acct.name == 'BankTwoStale'
        assert acct.description == 'Stale Bank Account'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path == 'secret/foo/bar/BankTwo'
        assert acct.ofxgetter_config_json == '{"foo": "baz"}'
        assert acct.negate_ofx_amounts is True
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Bank
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is True

    def test_21_get_acct2_click(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_xpath('//a[text()="BankTwoStale"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 2'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == '2'
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == 'BankTwoStale'
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == 'Stale Bank Account'
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected()
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == 'secret/foo/bar/BankTwo'
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == '{"foo": "baz"}'
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected() is True
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected() is True
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed() is False
        # END CREDIT
        assert selenium.find_element_by_id('account_frm_active').is_selected()

    def test_22_edit_acct2(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_xpath('//a[text()="BankTwoStale"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 2'
        selenium.find_element_by_id('account_frm_name').send_keys('Edited')
        selenium.find_element_by_id('account_frm_description').clear()
        selenium.find_element_by_id('account_frm_description').send_keys(
            'a2desc'
        )
        selenium.find_element_by_id('account_frm_ofx_cat_memo').click()
        selenium.find_element_by_id('account_frm_vault_creds_path').send_keys(
            '/baz'
        )
        selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).clear()
        selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).send_keys('{"key": "value"}')
        selenium.find_element_by_id('account_frm_negate_ofx').click()
        selenium.find_element_by_id('account_frm_reconcile_trans').click()
        selenium.find_element_by_id('account_frm_active').click()
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 2 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_23_verify_db(self, testdb):
        acct = testdb.query(Account).get(2)
        assert acct is not None
        assert acct.name == 'BankTwoStaleEdited'
        assert acct.description == 'a2desc'
        assert acct.ofx_cat_memo_to_name is True
        assert acct.vault_creds_path == 'secret/foo/bar/BankTwo/baz'
        assert acct.ofxgetter_config_json == '{"key": "value"}'
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is False
        assert acct.acct_type == AcctType.Bank
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is False

    def test_30_verify_db(self, testdb):
        acct = testdb.query(Account).get(3)
        assert acct is not None
        assert acct.name == 'CreditOne'
        assert acct.description == 'First Credit Card, limit 2000'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path is None
        assert acct.ofxgetter_config_json is None
        assert acct.negate_ofx_amounts is True
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Credit
        assert acct.credit_limit == Decimal('2000.0')
        assert acct.apr is None
        assert acct.prime_rate_margin == Decimal('0.005')
        assert acct.interest_class_name == 'AdbCompoundedDaily'
        assert acct.min_payment_class_name == 'MinPaymentAmEx'
        assert acct.is_active is True

    def test_31_get_acct3_click(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_xpath('//a[text()="CreditOne"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 3'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == '3'
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == 'CreditOne'
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == 'First Credit Card, limit 2000'
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected()
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected()
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected()
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed()
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').get_attribute('value') == '2000'
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed()
        assert selenium.find_element_by_id(
            'account_frm_apr').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed()
        assert selenium.find_element_by_id(
            'account_frm_margin').get_attribute('value') == '0.005'
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed()
        assert Select(selenium.find_element_by_id(
            'account_frm_int_class_name')
        ).first_selected_option.get_attribute('value') == 'AdbCompoundedDaily'
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed()
        assert Select(selenium.find_element_by_id(
            'account_frm_min_pay_class_name')
        ).first_selected_option.get_attribute('value') == 'MinPaymentAmEx'
        # END CREDIT
        assert selenium.find_element_by_id('account_frm_active').is_selected()

    def test_32_edit_acct3(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_xpath('//a[text()="CreditOne"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 3'
        selenium.find_element_by_id('account_frm_name').send_keys('Edited')
        selenium.find_element_by_id('account_frm_margin').clear()
        selenium.find_element_by_id('account_frm_margin').send_keys('4.21')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 3 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_33_verify_db(self, testdb):
        acct = testdb.query(Account).get(3)
        assert acct is not None
        assert acct.name == 'CreditOneEdited'
        assert acct.description == 'First Credit Card, limit 2000'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path is None
        assert acct.ofxgetter_config_json is None
        assert acct.negate_ofx_amounts is True
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Credit
        assert acct.credit_limit == Decimal('2000.0')
        assert acct.apr is None
        assert acct.prime_rate_margin == Decimal('0.0421')
        assert acct.interest_class_name == 'AdbCompoundedDaily'
        assert acct.min_payment_class_name == 'MinPaymentAmEx'
        assert acct.is_active is True

    def test_40_verify_db(self, testdb):
        acct = testdb.query(Account).get(4)
        assert acct is not None
        assert acct.name == 'CreditTwo'
        assert acct.description == 'Credit 2 limit 5500'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path == '/foo/bar'
        assert acct.ofxgetter_config_json == ''
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Credit
        assert acct.credit_limit == Decimal('5500.0')
        assert acct.apr == Decimal('0.1')
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name == 'AdbCompoundedDaily'
        assert acct.min_payment_class_name == 'MinPaymentDiscover'
        assert acct.is_active is True

    def test_41_get_acct4_url(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/4')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 4'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == '4'
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == 'CreditTwo'
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == 'Credit 2 limit 5500'
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected()
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == '/foo/bar'
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected()
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed()
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').get_attribute('value') == '5500'
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed()
        assert selenium.find_element_by_id(
            'account_frm_apr').get_attribute('value') == '0.1'
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed()
        assert selenium.find_element_by_id(
            'account_frm_margin').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed()
        assert Select(selenium.find_element_by_id(
            'account_frm_int_class_name')
        ).first_selected_option.get_attribute('value') == 'AdbCompoundedDaily'
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed()
        assert Select(selenium.find_element_by_id(
            'account_frm_min_pay_class_name')
        ).first_selected_option.get_attribute('value') == 'MinPaymentDiscover'
        # END CREDIT
        assert selenium.find_element_by_id('account_frm_active').is_selected()

    def test_42_edit_acct4(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts/4')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 4'
        selenium.find_element_by_id('account_frm_name').send_keys('Edited')
        selenium.find_element_by_id('account_frm_description').clear()
        selenium.find_element_by_id('account_frm_description').send_keys(
            'a4desc'
        )
        selenium.find_element_by_id('account_frm_ofx_cat_memo').click()
        selenium.find_element_by_id('account_frm_vault_creds_path').send_keys(
            '/baz'
        )
        selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).clear()
        selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).send_keys('{"key": "value"}')
        selenium.find_element_by_id('account_frm_negate_ofx').click()
        selenium.find_element_by_id('account_frm_reconcile_trans').click()
        # BEGIN CREDIT
        selenium.find_element_by_id('account_frm_credit_limit').clear()
        selenium.find_element_by_id('account_frm_credit_limit').send_keys(
            '123.45'
        )
        selenium.find_element_by_id('account_frm_apr').clear()
        selenium.find_element_by_id('account_frm_apr').send_keys('45.67')
        selenium.find_element_by_id('account_frm_margin').clear()
        Select(selenium.find_element_by_id(
            'account_frm_int_class_name')
        ).select_by_value('AdbCompoundedDaily')
        Select(selenium.find_element_by_id(
            'account_frm_min_pay_class_name')
        ).select_by_value('MinPaymentCiti')
        # END CREDIT
        selenium.find_element_by_id('account_frm_active').click()
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 4 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_43_verify_db(self, testdb):
        acct = testdb.query(Account).get(4)
        assert acct is not None
        assert acct.name == 'CreditTwoEdited'
        assert acct.description == 'a4desc'
        assert acct.ofx_cat_memo_to_name is True
        assert acct.vault_creds_path == '/foo/bar/baz'
        assert acct.ofxgetter_config_json == '{"key": "value"}'
        assert acct.negate_ofx_amounts is True
        assert acct.reconcile_trans is False
        assert acct.acct_type == AcctType.Credit
        assert acct.credit_limit == Decimal('123.45')
        assert acct.apr == Decimal('0.4567')
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name == 'AdbCompoundedDaily'
        assert acct.min_payment_class_name == 'MinPaymentCiti'
        assert acct.is_active is False

    def test_50_verify_db(self, testdb):
        acct = testdb.query(Account).get(5)
        assert acct is not None
        assert acct.name == 'InvestmentOne'
        assert acct.description == 'Investment One Stale'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path == ''
        assert acct.ofxgetter_config_json == ''
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is False
        assert acct.acct_type == AcctType.Investment
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is True

    def test_51_get_acct5_click(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_xpath('//a[text()="InvestmentOne"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 5'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == '5'
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == 'InvestmentOne'
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == 'Investment One Stale'
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected()
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected() is False
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed() is False
        # END CREDIT
        assert selenium.find_element_by_id('account_frm_active').is_selected()

    def test_52_edit_acct5(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_xpath('//a[text()="InvestmentOne"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Account 5'
        selenium.find_element_by_id('account_frm_name').send_keys('Edited')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 5 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_53_verify_db(self, testdb):
        acct = testdb.query(Account).get(5)
        assert acct is not None
        assert acct.name == 'InvestmentOneEdited'
        assert acct.description == 'Investment One Stale'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path is None
        assert acct.ofxgetter_config_json is None
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is False
        assert acct.acct_type == AcctType.Investment
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is True

    def test_60_verify_db(self, testdb):
        max_id = max([
            t.id for t in testdb.query(Account).all()
        ])
        assert max_id == 6

    def test_61_modal_add_bank(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_id('btn_add_acct_bank')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Account'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == ''
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == ''
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected() is True
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected()
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed() is False
        # END CREDIT
        assert selenium.find_element_by_id(
            'account_frm_active').is_selected() is True
        selenium.find_element_by_id('account_frm_name').send_keys('Acct7')
        selenium.find_element_by_id('account_frm_description').clear()
        selenium.find_element_by_id('account_frm_description').send_keys(
            'a7desc'
        )
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 7 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_62_verify_db(self, testdb):
        acct = testdb.query(Account).get(7)
        assert acct is not None
        assert acct.name == 'Acct7'
        assert acct.description == 'a7desc'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path is None
        assert acct.ofxgetter_config_json is None
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Bank
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is True

    def test_70_verify_db(self, testdb):
        max_id = max([
            t.id for t in testdb.query(Account).all()
        ])
        assert max_id == 7

    def test_71_modal_add_credit(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_id('btn_add_acct_credit')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Account'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == ''
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == ''
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected() is True
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected()
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed() is False
        # END CREDIT
        assert selenium.find_element_by_id(
            'account_frm_active').is_selected() is True
        selenium.find_element_by_id('account_frm_name').send_keys('Acct8')
        selenium.find_element_by_id('account_frm_description').clear()
        selenium.find_element_by_id('account_frm_description').send_keys(
            'a8desc'
        )
        selenium.find_element_by_id('account_frm_type_credit').click()
        selenium.find_element_by_id('account_frm_ofx_cat_memo').click()
        selenium.find_element_by_id('account_frm_vault_creds_path').send_keys(
            '/path/to/creds'
        )
        selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).send_keys('{"foo": "blam"}')
        selenium.find_element_by_id('account_frm_negate_ofx').click()
        selenium.find_element_by_id('account_frm_reconcile_trans').click()
        # BEGIN CREDIT
        selenium.find_element_by_id('account_frm_credit_limit').clear()
        selenium.find_element_by_id('account_frm_credit_limit').send_keys(
            '987.65'
        )
        selenium.find_element_by_id('account_frm_apr').clear()
        selenium.find_element_by_id('account_frm_apr').send_keys('45.67')
        selenium.find_element_by_id('account_frm_margin').clear()
        Select(selenium.find_element_by_id(
            'account_frm_int_class_name')
        ).select_by_value('AdbCompoundedDaily')
        Select(selenium.find_element_by_id(
            'account_frm_min_pay_class_name')
        ).select_by_value('MinPaymentCiti')
        # END CREDIT
        selenium.find_element_by_id('account_frm_active').click()
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 8 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_72_verify_db(self, testdb):
        acct = testdb.query(Account).get(8)
        assert acct is not None
        assert acct.name == 'Acct8'
        assert acct.description == 'a8desc'
        assert acct.ofx_cat_memo_to_name is True
        assert acct.vault_creds_path == '/path/to/creds'
        assert acct.ofxgetter_config_json == '{"foo": "blam"}'
        assert acct.negate_ofx_amounts is True
        assert acct.reconcile_trans is False
        assert acct.acct_type == AcctType.Credit
        assert acct.credit_limit == Decimal('987.65')
        assert acct.apr == Decimal('0.4567')
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name == 'AdbCompoundedDaily'
        assert acct.min_payment_class_name == 'MinPaymentCiti'
        assert acct.is_active is False

    def test_80_verify_db(self, testdb):
        max_id = max([
            t.id for t in testdb.query(Account).all()
        ])
        assert max_id == 8

    def test_81_modal_add_bank(self, base_url, selenium):
        self.get(selenium, base_url + '/accounts')
        link = selenium.find_element_by_id('btn_add_acct_invest')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Account'
        assert selenium.find_element_by_id('account_frm_id').get_attribute(
            'value') == ''
        assert selenium.find_element_by_id('account_frm_name').get_attribute(
            'value') == ''
        assert selenium.find_element_by_id(
            'account_frm_description'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_type_bank').is_selected() is True
        assert selenium.find_element_by_id(
            'account_frm_type_credit').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_type_investment').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_ofx_cat_memo').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_vault_creds_path'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_ofxgetter_config_json'
        ).get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'account_frm_negate_ofx').is_selected() is False
        assert selenium.find_element_by_id(
            'account_frm_reconcile_trans').is_selected()
        # BEGIN CREDIT
        assert selenium.find_element_by_id(
            'account_frm_credit_limit').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_apr').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_margin').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_int_class_name').is_displayed() is False
        assert selenium.find_element_by_id(
            'account_frm_min_pay_class_name').is_displayed() is False
        # END CREDIT
        assert selenium.find_element_by_id(
            'account_frm_active').is_selected() is True
        selenium.find_element_by_id('account_frm_name').send_keys('Acct9')
        selenium.find_element_by_id('account_frm_description').clear()
        selenium.find_element_by_id('account_frm_description').send_keys(
            'a9desc'
        )
        selenium.find_element_by_id('account_frm_type_investment').click()
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Account 9 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()

    def test_82_verify_db(self, testdb):
        acct = testdb.query(Account).get(9)
        assert acct is not None
        assert acct.name == 'Acct9'
        assert acct.description == 'a9desc'
        assert acct.ofx_cat_memo_to_name is False
        assert acct.vault_creds_path is None
        assert acct.ofxgetter_config_json is None
        assert acct.negate_ofx_amounts is False
        assert acct.reconcile_trans is True
        assert acct.acct_type == AcctType.Investment
        assert acct.credit_limit is None
        assert acct.apr is None
        assert acct.prime_rate_margin is None
        assert acct.interest_class_name is None
        assert acct.min_payment_class_name is None
        assert acct.is_active is True



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import os
import logging
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
import socket
import json
from time import time

import biweeklybudget.settings
from biweeklybudget.tests.fixtures.sampledata import SampleDataLoader
from biweeklybudget.tests.sqlhelpers import restore_mysqldump, do_mysqldump

try:
    from pytest_flask.fixtures import LiveServer
except ImportError:
    pass

connstr = os.environ.get('DB_CONNSTRING', None)
if connstr is None:
    connstr = 'mysql+pymysql://budgetTester:jew8fu0ue@127.0.0.1:3306/' \
              'budgettest?charset=utf8mb4'
    os.environ['DB_CONNSTRING'] = connstr
biweeklybudget.settings.DB_CONNSTRING = connstr

import biweeklybudget.db  # noqa
import biweeklybudget.models.base  # noqa
from biweeklybudget.db_event_handlers import init_event_listeners  # noqa
from biweeklybudget.tests.unit.test_interest import InterestData  # noqa

engine = create_engine(
    connstr, convert_unicode=True, echo=False,
    connect_args={'sql_mode': 'STRICT_ALL_TABLES'},
    pool_size=10, pool_timeout=120
)

logger = logging.getLogger(__name__)

# suppress webdriver DEBUG logging
selenium_log = logging.getLogger("selenium")
selenium_log.setLevel(logging.INFO)
selenium_log.propagate = True

class_refresh_db_durations = []


@pytest.fixture(scope='session')
def dump_file_path(tmpdir_factory):
    """
    Return the directory to use for the SQL dump files
    """
    return tmpdir_factory.mktemp('sqldump')


@pytest.fixture(scope="session")
def refreshdb(dump_file_path):
    """
    Refresh/Load DB data before tests; also exec mysqldump to write a
    SQL dump file for faster refreshes during test runs.
    """
    if 'NO_REFRESH_DB' not in os.environ:
        # setup the connection
        conn = engine.connect()
        logger.info('Refreshing DB (session-scoped)')
        # clean the database
        biweeklybudget.models.base.Base.metadata.drop_all(engine)
        biweeklybudget.models.base.Base.metadata.create_all(engine)
        # load the sample data
        data_sess = scoped_session(
            sessionmaker(autocommit=False, autoflush=False, bind=conn)
        )
        SampleDataLoader(data_sess).load()
        data_sess.flush()
        data_sess.commit()
        data_sess.close()
        # close connection
        conn.close()
    else:
        logger.info('Skipping session-scoped DB refresh')
    # write the dump files
    do_mysqldump(dump_file_path, engine)
    do_mysqldump(dump_file_path, engine, with_data=False)
    yield


@pytest.fixture(scope="class")
def class_refresh_db(dump_file_path):
    """
    This fixture rolls the DB back to the previous state when the class is
    finished; to be used on classes that alter data.

    Use like:

        @pytest.mark.usefixtures('class_refresh_db', 'testdb')
        class MyClass(AcceptanceHelper):
    """
    global class_refresh_db_durations
    yield
    if 'NO_CLASS_REFRESH_DB' in os.environ:
        return
    logger.info('Refreshing DB (class-scoped)')
    s = time()
    restore_mysqldump(dump_file_path, engine)
    class_refresh_db_durations.append(time() - s)


@pytest.fixture
def testdb():
    """
    DB fixture to be used in tests
    """
    # setup the connection
    conn = engine.connect()
    sess = sessionmaker(autocommit=False, bind=conn)()
    init_event_listeners(sess)
    # yield the session
    yield(sess)
    sess.close()
    conn.close()


@pytest.fixture(scope="session")
def testflask():
    """
    This is a version of pytest-flask's live_server fixture, modified for
    session use.
    """
    if 'BIWEEKLYBUDGET_TEST_BASE_URL' not in os.environ:
        # Bind to an open port
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('', 0))
        port = s.getsockname()[1]
        s.close()
        from biweeklybudget.flaskapp.app import app  # noqa
        server = LiveServer(app, port)
        server.start()
        yield(server)
        server.stop()


@pytest.fixture(scope="session")
def base_url(testflask):
    """
    Simple fixture to return ``testflask`` base URL
    """
    url = os.environ.get('BIWEEKLYBUDGET_TEST_BASE_URL', None)
    if url is not None:
        return url
    return testflask.url()


@pytest.fixture
def selenium(selenium):
    """
    Per pytest-selenium docs, use this to override the selenium fixture to
    provide global common setup.
    """
    selenium.set_window_size(1920, 1080)
    selenium.implicitly_wait(2)
    # from http://stackoverflow.com/a/13853684/211734
    selenium.set_script_timeout(30)
    # from http://stackoverflow.com/a/17536547/211734
    selenium.set_page_load_timeout(30)
    return selenium


@pytest.fixture
def chrome_options(chrome_options):
    chrome_options.add_argument('headless')
    return chrome_options


def pytest_addoption(parser):
    group = parser.getgroup("terminal reporting", "reporting", after="general")
    group.addoption('--durations-file',
                    action="store", type=str, default=None, metavar="P",
                    help="write all durations as JSON to P")


def pytest_terminal_summary(terminalreporter):
    """Write all test durations to results/durations.json"""
    fpath = terminalreporter.config.option.durations_file
    if fpath is None:
        return
    tr = terminalreporter
    dlist = []
    for replist in tr.stats.values():
        for rep in replist:
            if hasattr(rep, 'duration'):
                dlist.append(rep)
    if not dlist:
        return
    dlist.sort(key=lambda x: x.duration)
    dlist.reverse()
    tr.write_sep("-", "wrote all test durations to: %s" % fpath)

    result = []
    for rep in dlist:
        nodeid = rep.nodeid.replace("::()::", "::")
        result.append([nodeid, rep.when, rep.duration])
    with open(fpath, 'w') as fh:
        fh.write(json.dumps({
            'requests': result,
            'class_refresh_db': class_refresh_db_durations
        }))


# next section from:
# https://docs.pytest.org/en/latest/example/simple.html#\
# incremental-testing-test-steps


def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        if call.excinfo is not None:
            parent = item.parent
            parent._previousfailed = item


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        previousfailed = getattr(item.parent, "_previousfailed", None)
        if previousfailed is not None:
            pytest.xfail("previous test failed (%s)" % previousfailed.name)


"""
Begin generated/parametrized tests for interest calculation.

I REALLY wish pytest still supported yield tests, this is sooooo messy!
"""


def pytest_generate_tests(metafunc):
    if (
        metafunc.function.__name__.startswith('test_calculate') and
        metafunc.module.__name__ == 'biweeklybudget.tests.unit.test_interest'
    ):
        if metafunc.cls.__name__ == 'TestDataAmEx':
            param_for_adbdaily_calc(metafunc, InterestData.amex)
        if metafunc.cls.__name__ == 'TestDataCiti':
            param_for_adbdaily_calc(metafunc, InterestData.citi)
        if metafunc.cls.__name__ == 'TestDataDiscover':
            param_for_adbdaily_calc(metafunc, InterestData.discover)


def param_for_adbdaily_calc(metafunc, s):
    dates = [d['start'].strftime('%Y-%m-%d') for d in s]
    metafunc.parametrize(
        'data',
        s,
        ids=dates
    )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from datetime import timedelta, datetime, date
from functools import total_ordering
from sqlalchemy import or_, asc
from dateutil import relativedelta
from collections import defaultdict
from decimal import Decimal

from biweeklybudget import settings
from biweeklybudget.models import Transaction, ScheduledTransaction, Budget
from biweeklybudget.utils import dtnow


@total_ordering
class BiweeklyPayPeriod(object):
    """
    This object contains all logic related to working with pay periods,
    specifically finding a pay period for a given data, and figuring out the
    start and end dates of pay periods. Sure, the app is called "biweeklybudget"
    but there's no reason to hard-code logic all over the place that's this
    simple.
    """

    def __init__(self, start_date, db_session):
        """
        Create a new BiweeklyPayPeriod instance.

        :param start_date: starting date of the pay period
        :type start_date: :py:class:`datetime.date` or
          :py:class:`datetime.datetime`
        :param db_session: active database session to use for queries
        :type db_session: sqlalchemy.orm.session.Session
        """
        if isinstance(start_date, datetime):
            start_date = start_date.date()
        self._db = db_session
        self._start_date = start_date
        self._end_date = start_date + self.period_length
        self._data_cache = {}
        self._income_budget_id_list = None

    @property
    def period_interval(self):
        """
        Return the interval between BiweeklyPayPeriods as a timedelta.

        :return: interval between BiweeklyPayPeriods
        :rtype: datetime.timedelta
        """
        return timedelta(days=14)

    @property
    def period_length(self):
        """
        Return the length of a BiweeklyPayPeriod; this is calculated as
        :py:attr:`~.period_interval` minus one second.

        :return: length of one BiweeklyPayPeriod
        :rtype: datetime.timedelta
        """
        return self.period_interval - timedelta(days=1)

    @property
    def start_date(self):
        """
        Return the starting date for this pay period. The period is generally
        considered to start at midnight (00:00) of this date.

        :return: start date for pay period
        :rtype: datetime.date
        """
        return self._start_date

    @property
    def end_date(self):
        """
        Return the date of the last day in this pay period. The pay period is
        generally considered to end at the last instant (i.e. 23:59:59) of this
        date.

        :return: last date in the pay period
        :rtype: datetime.date
        """
        return self._end_date

    @property
    def is_in_past(self):
        return self.end_date < dtnow().date()

    @property
    def next(self):
        """
        Return the BiweeklyPayPeriod following this one.

        :return: next BiweeklyPayPeriod after this one
        :rtype: BiweeklyPayPeriod
        """
        return BiweeklyPayPeriod(
            (self.start_date + self.period_interval),
            self._db
        )

    @property
    def previous(self):
        """
        Return the BiweeklyPayPeriod preceding this one.

        :return: previous BiweeklyPayPeriod before this one
        :rtype: BiweeklyPayPeriod
        """
        return BiweeklyPayPeriod(
            (self.start_date - self.period_interval),
            self._db
        )

    def __repr__(self):
        return '<BiweeklyPayPeriod(%s)>' % self._start_date.strftime('%Y-%m-%d')

    def __eq__(self, other):
        if not isinstance(other, BiweeklyPayPeriod):
            return NotImplemented
        return self.start_date == other.start_date

    def __lt__(self, other):
        if not isinstance(other, BiweeklyPayPeriod):
            return NotImplemented
        return self.start_date < other.start_date

    @staticmethod
    def period_for_date(dt, db_session):
        """
        Given a datetime, return the BiweeklyPayPeriod instance describing the
        pay period containing this date.

        .. todo:: This is a very naive, poorly-performing implementation.

        :param dt: datetime or date to find the pay period for
        :type dt: :py:class:`~datetime.datetime` or :py:class:`~datetime.date`
        :param db_session: active database session to use for queries
        :type db_session: sqlalchemy.orm.session.Session
        :return: BiweeklyPayPeriod containing the specified date
        :rtype: :py:class:`~.BiweeklyPayPeriod`
        """
        p = BiweeklyPayPeriod(settings.PAY_PERIOD_START_DATE, db_session)
        if isinstance(dt, datetime):
            dt = dt.date()
        if dt < p.start_date:
            while True:
                if p.end_date >= dt >= p.start_date:
                    return p
                p = p.previous
        if dt > p.end_date:
            while True:
                if p.end_date >= dt >= p.start_date:
                    return p
                p = p.next
        return p

    def filter_query(self, query, date_prop):
        """
        Filter ``query`` for ``date_prop`` in this pay period. Returns a copy
        of the query.

        e.g. to filter an existing query of :py:class:`~.OFXTransaction` for
        the BiweeklyPayPeriod starting on 2017-01-14:

        .. code-block:: python

            q = # some query here
            p = BiweeklyPayPeriod(date(2017, 1, 14))
            q = p.filter_query(q, OFXTransaction.date_posted)

        :param query: The query to filter
        :type query: ``sqlalchemy.orm.query.Query``
        :param date_prop: the Model's date property, to filter on.
        :return: the filtered query
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        return query.filter(
            date_prop >= self.start_date, date_prop <= self.end_date
        )

    def _transactions(self):
        """
        Return a Query for all :py:class:`~.Transaction` for this pay period.

        :return: Query matching all Transactions for this pay period
        :rtype: sqlalchemy.orm.query.Query
        """
        return self.filter_query(
            self._db.query(Transaction),
            Transaction.date
        )

    def _scheduled_transactions_date(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        date (schedule_type == "date") for this pay period.

        :return: Query matching all ScheduledTransactions defined by date, for
          this pay period.
        :rtype: sqlalchemy.orm.query.Query
        """
        return self.filter_query(
            self._db.query(ScheduledTransaction).filter(
                ScheduledTransaction.is_active.__eq__(True)),
            ScheduledTransaction.date
        )

    def _scheduled_transactions_per_period(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        number per period (schedule_type == "per period") for this pay period.

        :return: Query matching all ScheduledTransactions defined by number
          per period, for this pay period.
        :rtype: sqlalchemy.orm.query.Query
        """
        return self._db.query(ScheduledTransaction).filter(
            ScheduledTransaction.schedule_type.__eq__('per period'),
            ScheduledTransaction.is_active.__eq__(True)
        ).order_by(
            asc(ScheduledTransaction.num_per_period),
            asc(ScheduledTransaction.amount)
        )

    def _scheduled_transactions_monthly(self):
        """
        Return a Query for all :py:class:`~.ScheduledTransaction` defined by
        day of month (schedule_type == "monthly") for this pay period.

        :return: Query matching all ScheduledTransactions defined by day of
          month (monthly) for this period.
        :rtype: sqlalchemy.orm.query.Query
        """
        if self.start_date.day < self.end_date.day:
            # start and end dates are contiguous, in the same month
            return self._db.query(ScheduledTransaction).filter(
                ScheduledTransaction.schedule_type.__eq__('monthly'),
                ScheduledTransaction.is_active.__eq__(True),
                ScheduledTransaction.day_of_month <= self.end_date.day,
                ScheduledTransaction.day_of_month >= self.start_date.day
            )
        # else we span two months
        return self._db.query(ScheduledTransaction).filter(
            ScheduledTransaction.schedule_type.__eq__('monthly'),
            ScheduledTransaction.is_active.__eq__(True),
            or_(
                ScheduledTransaction.day_of_month <= self.end_date.day,
                ScheduledTransaction.day_of_month >= self.start_date.day
            )
        )

    @property
    def transactions_list(self):
        """
        Return an ordered list of dicts, each representing a transaction for
        this pay period. Dicts have keys and values as described in
        :py:meth:`~._trans_dict`.

        :return: ordered list of transaction dicts
        :rtype: list
        """
        return self._data['all_trans_list']

    @property
    def _income_budget_ids(self):
        """
        Return a list of all :py:class:`~.Budget` IDs for Income budgets.

        :return: list of income budget IDs
        :rtype: list
        """
        if self._income_budget_id_list is None:
            self._income_budget_id_list = [
                b.id for b in self._db.query(
                    Budget).filter(Budget.is_income.__eq__(True)).all()
            ]
        return self._income_budget_id_list

    @property
    def _data(self):
        """
        Return the object-local data cache dict. Built it if not already
        present.

        :return: object-local data cache
        :rtype: dict
        """
        if len(self._data_cache) > 0:
            return self._data_cache
        self._data_cache = {
            'transactions': self._transactions().all(),
            'st_date': self._scheduled_transactions_date().all(),
            'st_per_period': self._scheduled_transactions_per_period().all(),
            'st_monthly': self._scheduled_transactions_monthly().all()
        }
        self._data_cache['all_trans_list'] = self._make_combined_transactions()
        self._data_cache['budget_sums'] = self._make_budget_sums()
        self._data_cache['overall_sums'] = self._make_overall_sums()
        return self._data_cache

    def clear_cache(self):
        """
        Clear the cached transaction, budget and sum data stored in
        `self._data_cache` and returned by :py:attr:`~._data`.
        """
        self._data_cache = {}

    def _make_combined_transactions(self):
        """
        Combine all Transactions and ScheduledTransactions from
        ``self._data_cache`` into one ordered list of similar dicts, adding
        dates to the monthly ScheduledTransactions as appropriate and excluding
        ScheduledTransactions that have been converted to real Transactions.
        Store the finished list back into ``self._data_cache``.
        """
        unordered = []
        # ScheduledTransaction ID to count of real trans for each
        st_ids = defaultdict(int)
        for t in self._data_cache['transactions']:
            unordered.append(self._trans_dict(t))
            if t.scheduled_trans_id is not None:
                st_ids[t.scheduled_trans_id] += 1
        for t in self._data_cache['st_date']:
            if t.id not in st_ids:
                unordered.append(self._trans_dict(t))
        for t in self._data_cache['st_monthly']:
            if t.id not in st_ids:
                unordered.append(self._trans_dict(t))
        ordered = []
        for t in self._data_cache['st_per_period']:
            d = self._trans_dict(t)
            for _ in range(0, (t.num_per_period - st_ids[t.id])):
                ordered.append(d)
        for t in sorted(unordered, key=lambda k: k['date']):
            ordered.append(t)

        def sortkey(k):
            d = k.get('date', None)
            if d is None:
                d = date.min
            return d, k['amount']

        return sorted(ordered, key=sortkey)

    @property
    def budget_sums(self):
        """
        Return a dict of budget sums; the return value of
        :py:meth:`~._make_budget_sums`.

        :return: dict of dicts, transaction sums and amounts per budget
        :rtype: dict
        """
        return self._data['budget_sums']

    def _make_budget_sums(self):
        """
        Find the sums of all transactions per periodic budget ID ; return a dict
        where keys are budget IDs and values are per-budget dicts containing:

        - ``budget_amount`` *(Decimal.decimal)* - the periodic budget
          :py:attr:`~.Budget.starting_balance`.
        - ``allocated`` *(Decimal.decimal)* - sum of all
          :py:class:`~.ScheduledTransaction` and :py:class:`~.Transaction`
          amounts against the budget this period. For actual transactions, we
          use the :py:attr:`~.Transaction.budgeted_amount` if present (not
          None).
        - ``spent`` *(Decimal.decimal)* - the sum of all actual
          :py:class:`~.Transaction` amounts against the budget this period.
        - ``trans_total`` *(Decimal.decimal)* - the sum of spent amounts for
          Transactions that have them, or allocated amounts for
          ScheduledTransactions.
        - ``remaining`` *(Decimal.decimal)* - the remaining amount in the
          budget. This is ``budget_amount`` minus the greater of ``allocated``
          or ``trans_total``. For income budgets, this is always positive.

        :return: dict of dicts, transaction sums and amounts per budget
        :rtype: dict
        """
        res = {}
        for b in self._db.query(Budget).filter(
            Budget.is_active.__eq__(True),
            Budget.is_periodic.__eq__(True)
        ).all():
            res[b.id] = {
                'budget_amount': b.starting_balance,
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': b.is_income
            }
        for t in self.transactions_list:
            if t['budget_id'] not in res:
                # inactive budget
                continue
            if t['type'] == 'ScheduledTransaction':
                res[t['budget_id']]['allocated'] += t['amount']
                res[t['budget_id']]['trans_total'] += t['amount']
                continue
            # t['type'] == 'Transaction'
            res[t['budget_id']]['trans_total'] += t['amount']
            if t['budgeted_amount'] is None:
                res[t['budget_id']]['allocated'] += t['amount']
                res[t['budget_id']]['spent'] += t['amount']
            else:
                res[t['budget_id']]['allocated'] += t['budgeted_amount']
                res[t['budget_id']]['spent'] += t['amount']
        for b in res.keys():
            if res[b]['trans_total'] > res[b]['allocated']:
                res[b]['remaining'] = res[
                    b]['budget_amount'] - res[b]['trans_total']
            else:
                res[b]['remaining'] = res[
                    b]['budget_amount'] - res[b]['allocated']
            if res[b]['is_income']:
                res[b]['remaining'] = abs(res[b]['remaining'])
        return res

    @property
    def overall_sums(self):
        """
        Return a dict of overall sums; the return value of
        :py:meth:`~._make_overall_sums`.

        :return: dict describing sums for the pay period
        :rtype: dict
        """
        return self._data['overall_sums']

    def _make_overall_sums(self):
        """
        Return a dict describing the overall sums for this pay period, namely:

        - ``allocated`` *(Decimal.decimal)* total amount allocated via
          :py:class:`~.ScheduledTransaction`,
          :py:class:`~.Transaction` (counting the
          :py:attr:`~.Transaction.budgeted_amount` for Transactions that have
          one), or :py:class:`~.Budget` (not counting income budgets).
        - ``spent`` *(Decimal.decimal)* total amount actually spent via
          :py:class:`~.Transaction`.
        - ``income`` *(Decimal.decimal)* total amount of income allocated this
          pay period. Calculated value (from :py:meth:`~._make_budget_sums` /
          ``self._data_cache['budget_sums']``) should be negative, but is
          returned as its positive inverse (absolute value).
        - ``remaining`` *(Decimal.decimal)* income minus the greater of
          ``allocated`` or ``spent`` for current or future pay periods, or minus
          ``spent`` for pay periods ending in the past (:py:attr:`~.is_in_past`)

        :return: dict describing sums for the pay period
        :rtype: dict
        """
        res = {
            'allocated': Decimal('0.0'),
            'spent': Decimal('0.0'),
            'income': Decimal('0.0'),
            'remaining': Decimal('0.0')
        }
        for _, b in self._data_cache['budget_sums'].items():
            if b['is_income']:
                if abs(b['trans_total']) > abs(b['budget_amount']):
                    res['income'] += abs(b['trans_total'])
                else:
                    res['income'] += abs(b['budget_amount'])
                continue
            if b['allocated'] > b['budget_amount']:
                res['allocated'] += b['allocated']
            else:
                res['allocated'] += b['budget_amount']
            res['spent'] += b['spent']
        if res['spent'] > res['allocated'] or self.is_in_past:
            res['remaining'] = res['income'] - res['spent']
        else:
            res['remaining'] = res['income'] - res['allocated']
        return res

    def _trans_dict(self, t):
        """
        Given a Transaction or ScheduledTransaction, return a dict of a
        common format describing the object.

        The resulting dict will have the following layout:

        * ``type`` (**str**) "Transaction" or "ScheduledTransaction"
        * ``id`` (**int**) the id of the object
        * ``date`` (**date**) the date of the transaction, or None for
          per-period ScheduledTransactions
        * ``sched_type`` (**str**) for ScheduledTransactions, the schedule
          type ("monthly", "date", or "per period")
        * ``sched_trans_id`` (**int**) for Transactions, the
          ScheduledTransaction ``id`` that it was created from, or None.
        * ``description`` (**str**) the transaction description
        * ``amount`` (**Decimal.decimal**) the transaction amount
        * ``budgeted_amount`` (**Decimal.decimal**) the budgeted amount.
          This may be None.
        * ``account_id`` (**int**) the id of the Account the transaction is
          against.
        * ``account_name`` (**str**) the name of the Account the transaction is
          against.
        * ``budget_id`` (**int**) the id of the Budget the transaction is
          against.
        * ``budget_name`` (**str**) the name of the Budget the transaction is
          against.
        * ``reconcile_id`` (**int**) the ID of the TxnReconcile, or None

        :param t: the object to return a dict for
        :type t: :py:class:`~.Transaction` or :py:class:`~.ScheduledTransaction`
        :return: dict describing ``t``
        :rtype: dict
        """
        if isinstance(t, Transaction):
            return self._dict_for_trans(t)
        return self._dict_for_sched_trans(t)

    def _dict_for_trans(self, t):
        """
        Return a dict describing the Transaction t. Called from
        :py:meth:`~._trans_dict`.

        The resulting dict will have the following layout:

        * ``type`` (**str**) "Transaction" or "ScheduledTransaction"
        * ``id`` (**int**) the id of the object
        * ``date`` (**date**) the date of the transaction, or None for
          per-period ScheduledTransactions
        * ``sched_type`` (**str**) for ScheduledTransactions, the schedule
          type ("monthly", "date", or "per period")
        * ``sched_trans_id`` (**int**) for Transactions, the
          ScheduledTransaction ``id`` that it was created from, or None.
        * ``description`` (**str**) the transaction description
        * ``amount`` (**Decimal.decimal**) the transaction amount
        * ``budgeted_amount`` (**Decimal.decimal**) the budgeted amount.
          This may be None.
        * ``account_id`` (**int**) the id of the Account the transaction is
          against.
        * ``account_name`` (**str**) the name of the Account the transaction is
          against.
        * ``budget_id`` (**int**) the id of the Budget the transaction is
          against.
        * ``budget_name`` (**str**) the name of the Budget the transaction is
          against.
        * ``reconcile_id`` (**int**) the ID of the TxnReconcile, or None

        :param t: transaction to describe
        :type t: Transaction
        :return: common-format dict describing ``t``
        :rtype: dict
        """
        res = {
            'type': 'Transaction',
            'id': t.id,
            'date': t.date,
            'sched_type': None,
            'sched_trans_id': t.scheduled_trans_id,
            'description': t.description,
            'amount': t.actual_amount,
            'account_id': t.account_id,
            'account_name': t.account.name,
            'budget_id': t.budget_id,
            'budget_name': t.budget.name
        }
        if t.reconcile is None:
            res['reconcile_id'] = None
        else:
            res['reconcile_id'] = t.reconcile.id
        if t.budgeted_amount is None:
            res['budgeted_amount'] = None
        else:
            res['budgeted_amount'] = t.budgeted_amount
        return res

    def _dict_for_sched_trans(self, t):
        """
        Return a dict describing the ScheduledTransaction t. Called from
        :py:meth:`~._trans_dict`.

        The resulting dict will have the following layout:

        * ``type`` (**str**) "Transaction" or "ScheduledTransaction"
        * ``id`` (**int**) the id of the object
        * ``date`` (**date**) the date of the transaction, or None for
          per-period ScheduledTransactions
        * ``sched_type`` (**str**) for ScheduledTransactions, the schedule
          type ("monthly", "date", or "per period")
        * ``sched_trans_id`` **None**
        * ``description`` (**str**) the transaction description
        * ``amount`` (**Decimal.decimal**) the transaction amount
        * ``budgeted_amount`` **None**
        * ``account_id`` (**int**) the id of the Account the transaction is
          against.
        * ``account_name`` (**str**) the name of the Account the transaction is
          against.
        * ``budget_id`` (**int**) the id of the Budget the transaction is
          against.
        * ``budget_name`` (**str**) the name of the Budget the transaction is
          against.
        * ``reconcile_id`` (**int**) the ID of the TxnReconcile, or None

        :param t: ScheduledTransaction to describe
        :type t: ScheduledTransaction
        :return: common-format dict describing ``t``
        :rtype: dict
        """
        res = {
            'type': 'ScheduledTransaction',
            'id': t.id,
            'sched_type': t.schedule_type,
            'sched_trans_id': None,
            'description': t.description,
            'amount': t.amount,
            'budgeted_amount': None,
            'account_id': t.account_id,
            'account_name': t.account.name,
            'budget_id': t.budget_id,
            'budget_name': t.budget.name,
            'reconcile_id': None
        }
        if t.schedule_type == 'date':
            res['date'] = t.date
            return res
        if t.schedule_type == 'per period':
            res['date'] = None
            return res
        # else it's a monthly transaction, and we need to figure out the date
        # for it that falls in this PayPeriod.
        if self.start_date.day <= t.day_of_month <= self.end_date.day:
            res['date'] = date(
                year=self.start_date.year,
                month=self.start_date.month,
                day=t.day_of_month
            )
            return res
        # If we got here, we're in a pay period that spans two months; i.e.
        # start_date.day > end_date.day.
        if t.day_of_month >= self.start_date.day:
            # in the same month as start_date.day
            res['date'] = date(
                year=self.start_date.year,
                month=self.start_date.month,
                day=t.day_of_month
            )
            return res
        # else t.day_of_month < self.start_date.day, which means it's actually
        # ``t.day_of_month`` in the next month...
        res['date'] = date(
            year=self.start_date.year,
            month=self.start_date.month,
            day=t.day_of_month
        ) + relativedelta.relativedelta(months=1)
        return res



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""
import sys
from datetime import date
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import null

from biweeklybudget.models.transaction import Transaction
from biweeklybudget.tests.unit_helpers import binexp_to_dict

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import Mock, patch, call  # noqa
else:
    from unittest.mock import Mock, patch, call  # noqa

pbm = 'biweeklybudget.models.transaction'


class TestTransactionModel(object):

    @patch('%s.RECONCILE_BEGIN_DATE' % pbm, date(2017, 3, 17))
    def test_unreconciled(self):
        Transaction()
        m_db = Mock()
        m_q = Mock(spec_set=Query)
        m_filt = Mock(spec_set=Query)
        m_db.query.return_value = m_q
        m_q.filter.return_value = m_filt
        res = Transaction.unreconciled(m_db)
        assert res == m_filt
        assert len(m_db.mock_calls) == 2
        assert m_db.mock_calls[0] == call.query(Transaction)
        kall = m_db.mock_calls[1]
        assert kall[0] == 'query().filter'
        expected1 = Transaction.reconcile.__eq__(null())
        expected2 = Transaction.date.__ge__(date(2017, 3, 17))
        expected3 = Transaction.account.has(reconcile_trans=True)
        assert len(kall[1]) == 3
        assert str(expected1) == str(kall[1][0])
        assert binexp_to_dict(expected2) == binexp_to_dict(kall[1][1])
        assert str(expected3) == str(kall[1][2])



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import timedelta, date, datetime
from pytz import UTC
from selenium.webdriver.support.ui import Select
from decimal import Decimal

from biweeklybudget.utils import dtnow
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.txn_reconcile import TxnReconcile
from biweeklybudget.models.budget_model import Budget


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestTransactions(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'Transactions - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestTransactionsDefault(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.dt = dtnow()
        self.get(selenium, base_url + '/transactions')

    def test_table(self, selenium):
        table = selenium.find_element_by_id('table-transactions')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        assert texts == [
            [
                (self.dt + timedelta(days=4)).date().strftime('%Y-%m-%d'),
                '$111.13',
                'T1foo',
                'BankOne (1)',
                'Periodic1 (1)',
                'Yes (1)',
                '$111.11',
                'Yes (1)'
            ],
            [
                self.dt.date().strftime('%Y-%m-%d'),
                '-$333.33',
                'T2',
                'BankTwoStale (2)',
                'Standing1 (4)',
                'Yes (3)',
                '',
                ''
            ],
            [
                (self.dt - timedelta(days=2)).date().strftime('%Y-%m-%d'),
                '$222.22',
                'T3',
                'CreditOne (3)',
                'Periodic2 (2)',
                '',
                '',
                ''
            ]
        ]
        linkcols = [
            [
                c[2].get_attribute('innerHTML'),
                c[3].get_attribute('innerHTML'),
                c[4].get_attribute('innerHTML'),
                c[5].get_attribute('innerHTML'),
                c[7].get_attribute('innerHTML')
            ]
            for c in elems
        ]
        assert linkcols[0] == [
            '<a href="javascript:transModal(1, mytable)">T1foo</a>',
            '<a href="/accounts/1">BankOne (1)</a>',
            '<a href="/budgets/1">Periodic1 (1)</a>',
            '<a href="/scheduled/1">Yes (1)</a>',
            '<a href="javascript:txnReconcileModal(1)">Yes (1)</a>'
        ]
        assert linkcols[1] == [
            '<a href="javascript:transModal(2, mytable)">T2</a>',
            '<a href="/accounts/2">BankTwoStale (2)</a>',
            '<a href="/budgets/4">Standing1 (4)</a>',
            '<a href="/scheduled/3">Yes (3)</a>',
            '&nbsp;'
        ]
        assert linkcols[2] == [
            '<a href="javascript:transModal(3, mytable)">T3</a>',
            '<a href="/accounts/3">CreditOne (3)</a>',
            '<a href="/budgets/2">Periodic2 (2)</a>',
            '&nbsp;',
            '&nbsp;'
        ]

    def test_acct_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/transactions')
        acct_filter = Select(selenium.find_element_by_id('account_filter'))
        # find the options
        opts = []
        for o in acct_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]

    def test_acct_filter(self, selenium):
        p1trans = [
            'T1foo',
            'T2',
            'T3'
        ]
        self.get(selenium, self.baseurl + '/transactions')
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        # check sanity
        assert trans == p1trans
        acct_filter = Select(selenium.find_element_by_id('account_filter'))
        # select BankOne (1)
        acct_filter.select_by_value('1')
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == ['T1foo']
        # select back to all
        acct_filter.select_by_value('None')
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == p1trans

    def test_budg_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/transactions')
        budg_filter = Select(selenium.find_element_by_id('budget_filter'))
        # find the options
        opts = []
        for o in budg_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive']
        ]

    def test_budg_filter(self, selenium):
        p1trans = [
            'T1foo',
            'T2',
            'T3'
        ]
        self.get(selenium, self.baseurl + '/transactions')
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        # check sanity
        assert trans == p1trans
        budg_filter = Select(selenium.find_element_by_id('budget_filter'))
        # select Periodic2 (2)
        budg_filter.select_by_value('2')
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == ['T3']
        # select Standing1 (4)
        budg_filter.select_by_value('4')
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == ['T2']
        # select back to all
        budg_filter.select_by_value('None')
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        assert trans == p1trans

    def test_search(self, selenium):
        self.get(selenium, self.baseurl + '/transactions')
        search = self.retry_stale(
            selenium.find_element_by_xpath,
            '//input[@type="search"]'
        )
        search.send_keys('foo')
        self.wait_for_jquery_done(selenium)
        table = self.retry_stale(
            selenium.find_element_by_id,
            'table-transactions'
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[2] for t in texts]
        # check sanity
        assert trans == ['T1foo']


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestTransModalByURL(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(Transaction).get(3)
        assert t is not None
        assert t.description == 'T3'
        assert t.date == (dtnow() - timedelta(days=2)).date()
        assert float(t.actual_amount) == 222.22
        assert t.budgeted_amount is None
        assert t.account_id == 3
        assert t.budget_id == 2
        assert t.scheduled_trans_id is None
        assert t.notes == 'notesT3'

    def test_1_modal(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions/3')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 3'
        assert body.find_element_by_id(
            'trans_frm_id').get_attribute('value') == '3'
        assert body.find_element_by_id(
            'trans_frm_date').get_attribute('value') == (
            dtnow() - timedelta(days=2)).date().strftime('%Y-%m-%d')
        assert body.find_element_by_id(
            'trans_frm_amount').get_attribute('value') == '222.22'
        assert body.find_element_by_id(
            'trans_frm_description').get_attribute('value') == 'T3'
        acct_sel = Select(body.find_element_by_id('trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '3'
        budget_sel = Select(body.find_element_by_id('trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '2'
        assert selenium.find_element_by_id(
            'trans_frm_notes').get_attribute('value') == 'notesT3'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestTransModal(AcceptanceHelper):

    def test_00_simple_modal_verify_db(self, testdb):
        t = testdb.query(Transaction).get(2)
        assert t is not None
        assert t.description == 'T2'
        assert t.date == dtnow().date()
        assert float(t.actual_amount) == -333.33
        assert float(t.budgeted_amount) == -333.33
        assert t.account_id == 2
        assert t.budget_id == 4
        assert t.scheduled_trans_id == 3
        assert t.notes == 'notesT2'
        assert testdb.query(Budget).get(4).current_balance == Decimal('1284.23')
        assert testdb.query(Budget).get(5).current_balance == Decimal('9482.29')

    def test_01_simple_modal_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_xpath('//a[text()="T2"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 2'
        assert body.find_element_by_id(
            'trans_frm_id').get_attribute('value') == '2'
        assert body.find_element_by_id(
            'trans_frm_date').get_attribute('value') == dtnow().date(
        ).strftime('%Y-%m-%d')
        assert body.find_element_by_id(
            'trans_frm_amount').get_attribute('value') == '-333.33'
        assert body.find_element_by_id(
            'trans_frm_description').get_attribute('value') == 'T2'
        acct_sel = Select(body.find_element_by_id('trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '2'
        budget_sel = Select(body.find_element_by_id('trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '4'
        assert selenium.find_element_by_id(
            'trans_frm_notes').get_attribute('value') == 'notesT2'

    def test_02_simple_modal_modal_edit(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_xpath('//a[text()="T2"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 2'
        assert body.find_element_by_id(
            'trans_frm_id').get_attribute('value') == '2'
        d = body.find_element_by_id('trans_frm_date')
        d.clear()
        d.send_keys(
            (dtnow() - timedelta(days=3)).date().strftime('%Y-%m-%d')
        )
        amt = body.find_element_by_id('trans_frm_amount')
        amt.clear()
        amt.send_keys('-123.45')
        desc = body.find_element_by_id('trans_frm_description')
        desc.send_keys('edited')
        acct_sel = Select(body.find_element_by_id('trans_frm_account'))
        acct_sel.select_by_value('4')
        budget_sel = Select(body.find_element_by_id('trans_frm_budget'))
        budget_sel.select_by_value('5')
        notes = selenium.find_element_by_id('trans_frm_notes')
        notes.send_keys('edited')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 2 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element_by_id('table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'T2edited' in texts

    def test_03_simple_modal_verify_db(self, testdb):
        t = testdb.query(Transaction).get(2)
        assert t is not None
        assert t.description == 'T2edited'
        assert t.date == (dtnow() - timedelta(days=3)).date()
        assert float(t.actual_amount) == -123.45
        assert float(t.budgeted_amount) == -333.33
        assert t.account_id == 4
        assert t.budget_id == 5
        assert t.scheduled_trans_id == 3
        assert t.notes == 'notesT2edited'
        assert testdb.query(Budget).get(4).current_balance == Decimal('1074.35')
        assert testdb.query(Budget).get(5).current_balance == Decimal('9482.29')

    def test_10_cant_edit_reconciled_verify_db(self, testdb):
        t = testdb.query(Transaction).get(1)
        assert t is not None
        assert t.description == 'T1foo'
        assert t.date == (dtnow() + timedelta(days=4)).date()
        assert float(t.actual_amount) == 111.13
        assert float(t.budgeted_amount) == 111.11
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.scheduled_trans_id == 1
        assert t.notes == 'notesT1'

    def test_11_cant_edit_reconciled_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_xpath('//a[text()="T1foo"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 1'
        assert body.find_element_by_id(
            'trans_frm_id').get_attribute('value') == '1'
        assert body.find_element_by_id(
            'trans_frm_date').get_attribute('value') == (
            dtnow() + timedelta(days=4)).date().strftime('%Y-%m-%d')
        assert body.find_element_by_id(
            'trans_frm_amount').get_attribute('value') == '111.13'
        assert body.find_element_by_id(
            'trans_frm_description').get_attribute('value') == 'T1foo'
        acct_sel = Select(body.find_element_by_id('trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        budget_sel = Select(body.find_element_by_id('trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert selenium.find_element_by_id(
            'trans_frm_notes').get_attribute('value') == 'notesT1'

    def test_12_cant_edit_reconciled_modal_edit(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_xpath('//a[text()="T1foo"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 1'
        assert body.find_element_by_id(
            'trans_frm_id').get_attribute('value') == '1'
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-danger' in x.get_attribute('class')
        assert x.text.strip() == 'Server Error: Transaction 1 is already ' \
                                 'reconciled; cannot be edited.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_jquery_done(selenium)

    def test_22_modal_add(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_id('btn_add_trans')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Transaction'
        date_input = body.find_element_by_id('trans_frm_date')
        # BEGIN select the 15th of this month from the popup
        dnow = dtnow()
        expected_date = date(year=dnow.year, month=dnow.month, day=15)
        date_input.click()
        date_number = body.find_element_by_xpath(
            '//td[@class="day" and text()="15"]'
        )
        date_number.click()
        assert date_input.get_attribute(
            'value') == expected_date.strftime('%Y-%m-%d')
        # END date chooser popup
        amt = body.find_element_by_id('trans_frm_amount')
        amt.clear()
        amt.send_keys('123.45')
        desc = body.find_element_by_id('trans_frm_description')
        desc.send_keys('NewTrans4')
        acct_sel = Select(body.find_element_by_id('trans_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element_by_id('trans_frm_budget'))
        budget_sel.select_by_value('2')
        notes = selenium.find_element_by_id('trans_frm_notes')
        notes.send_keys('NewTransNotes')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 4 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that new trans was added to the table
        table = selenium.find_element_by_id('table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'NewTrans4' in texts

    def test_23_modal_add_verify_db(self, testdb):
        t = testdb.query(Transaction).get(4)
        assert t is not None
        assert t.description == 'NewTrans4'
        dnow = dtnow()
        assert t.date == date(year=dnow.year, month=dnow.month, day=15)
        assert float(t.actual_amount) == 123.45
        assert t.budgeted_amount is None
        assert t.account_id == 1
        assert t.budget_id == 2
        assert t.scheduled_trans_id is None
        assert t.notes == 'NewTransNotes'

    def test_31_verify_index_budgets_table(self, base_url, selenium):
        self.get(selenium, base_url + '/')
        stable = selenium.find_element_by_id('table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts == [
            ['Standing1 (4)', '$1,074.35'],
            ['Standing2 (5)', '$9,482.29']
        ]

    def test_32_modal_add(self, base_url, selenium):
        """Test that updating a transaction against a standing budget actually
            updates the balance on the standing budget."""
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_id('btn_add_trans')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Transaction'
        date_input = body.find_element_by_id('trans_frm_date')
        assert date_input.get_attribute(
            'value') == dtnow().strftime('%Y-%m-%d')
        # END date chooser popup
        amt = body.find_element_by_id('trans_frm_amount')
        amt.clear()
        amt.send_keys('345.67')
        desc = body.find_element_by_id('trans_frm_description')
        desc.send_keys('NewTrans5')
        acct_sel = Select(body.find_element_by_id('trans_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element_by_id('trans_frm_budget'))
        budget_sel.select_by_value('5')
        notes = selenium.find_element_by_id('trans_frm_notes')
        notes.send_keys('NewTransNotes')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 5 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that new trans was added to the table
        table = selenium.find_element_by_id('table-transactions')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'NewTrans5' in texts

    def test_33_verify_db(self, testdb):
        t = testdb.query(Transaction).get(5)
        assert t is not None
        assert t.description == 'NewTrans5'
        assert t.date == dtnow().date()
        assert float(t.actual_amount) == 345.67
        assert t.budgeted_amount is None
        assert t.account_id == 1
        assert t.budget_id == 5
        assert t.scheduled_trans_id is None
        assert t.notes == 'NewTransNotes'

    def test_34_verify_index_budgets_table(self, base_url, selenium):
        self.get(selenium, base_url + '/')
        stable = selenium.find_element_by_id('table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts == [
            ['Standing1 (4)', '$1,074.35'],
            ['Standing2 (5)', '$9,136.62']
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
@pytest.mark.incremental
class TestTransReconciledModal(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(TxnReconcile).get(1)
        assert t.ofx_account_id == 1
        assert t.ofx_fitid == 'BankOne.0.1'
        assert t.txn_id == 1
        assert t.rule_id is None
        assert t.note == 'reconcile notes'
        assert t.reconciled_at == datetime(2017, 4, 10, 8, 9, 11, tzinfo=UTC)

    def test_1_modal(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_xpath(
            '//a[@href="javascript:txnReconcileModal(1)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Transaction Reconcile 1'
        dl = body.find_element_by_tag_name('dl')
        assert dl.get_attribute('innerHTML') == '\n' \
            '<dt>Date Reconciled</dt><dd>2017-04-10 08:09:11 UTC</dd>\n' \
            '<dt>Note</dt><dd>reconcile notes</dd>\n' \
            '<dt>Rule</dt><dd>null</dd>\n'
        trans_tbl = body.find_element_by_id('txnReconcileModal-trans')
        trans_texts = self.tbody2textlist(trans_tbl)
        assert trans_texts == [
            ['Transaction'],
            [
                'Date',
                (dtnow() + timedelta(days=4)).strftime('%Y-%m-%d')
            ],
            ['Amount', '$111.13'],
            ['Budgeted Amount', '$111.11'],
            ['Description', 'T1foo'],
            ['Account', 'BankOne (1)'],
            ['Budget', 'Periodic1 (1)'],
            ['Notes', 'notesT1'],
            ['Scheduled?', 'Yes (1)']
        ]
        trans_elems = self.tbody2elemlist(trans_tbl)
        assert trans_elems[5][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'
        assert trans_elems[6][1].get_attribute('innerHTML') == '<a href=' \
            '"/budgets/1">Periodic1 (1)</a>'
        assert trans_elems[8][1].get_attribute('innerHTML') == '<a href=' \
            '"/scheduled/1">Yes (1)</a>'
        ofx_tbl = body.find_element_by_id('txnReconcileModal-ofx')
        ofx_texts = self.tbody2textlist(ofx_tbl)
        assert ofx_texts == [
            ['OFX Transaction'],
            ['Account', 'BankOne (1)'],
            ['FITID', 'BankOne.0.1'],
            ['Date Posted', (dtnow() - timedelta(days=6)).strftime('%Y-%m-%d')],
            ['Amount', '-$20.00'],
            ['Name', 'Late Fee'],
            ['Memo', ''],
            ['Type', 'Debit'],
            ['Description', ''],
            ['Notes', ''],
            ['Checknum', ''],
            ['MCC', ''],
            ['SIC', ''],
            ['OFX Statement'],
            ['ID', '1'],
            ['Date', (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')],
            ['Filename', '/stmt/BankOne/0'],
            [
                'File mtime',
                (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')
            ],
            ['Ledger Balance', '$12,345.67']
        ]
        ofx_elems = self.tbody2elemlist(ofx_tbl)
        assert ofx_elems[1][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'



from __future__ import unicode_literals
import logging
import sys
import argparse

from captain import echo, exit as console, ArgError
from captain.decorators import arg, args

from wishlist import __version__
from wishlist.core import Wishlist, ParseError, RobotError
from wishlist.browser import RecoverableCrash


# https://hg.python.org/cpython/file/2.7/Lib/argparse.py#l863
class LoggingAction(argparse.Action):
    def __init__(self, option_strings, dest, help=None, **kwargs):
        super(LoggingAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=True,
            default=False,
            required=False,
            help=help
        )

    def __call__(self, parser, namespace, values, option_string=None):
        """This is called if the value is actually passed in"""
        #pout.v(namespace, values, option_string)
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        log_handler = logging.StreamHandler(stream=sys.stderr)
        log_formatter = logging.Formatter('[%(levelname)s] %(message)s')
        log_handler.setFormatter(log_formatter)
        logger.addHandler(log_handler)
        setattr(namespace, self.dest, self.const)


def main_auth():
    """Signin to amazon so you can access private wishlists"""
    w = Wishlist()
    with w.open_full() as b:
        host = w.host
        echo.out("Requesting {}", host)
        b.location(host, ignore_cookies=True)

        # If you access from another country, amazon might prompt to redirect to
        # country specific store, we don't want that
        if b.element_exists("#redir-opt-out"):
            echo.out("Circumventing redirect")
            remember = b.element("#redir-opt-out")
            stay = b.element("#redir-stay-at-www")
            remember.click()
            stay.click()

        button = b.element("#a-autoid-0-announce")
        echo.out("Clicking sign in button")
        button.click()

        # now put in your creds
        email = b.element("#ap_email")
        password = b.element("#ap_password")
        submit = b.element("#signInSubmit")
        if email and password and submit:
            echo.out("Found sign in form")
            email_in = echo.prompt("Amazon email address")
            password_in = echo.prompt("Amazon password")

        email.send_keys(email_in)
        password.send_keys(password_in)
        echo.out("Signing in")
        submit.click()

        # for 2-factor, wait for this element
        code = b.wait_for_element("#auth-mfa-otpcode", 5)
        if code:
            echo.out("2-Factor authentication is on, you should be receiving a text")
            submit = b.element("#auth-signin-button")
            remember = b.element("#auth-mfa-remember-device")
            remember.click()
            authcode = echo.prompt("2-Factor authcode")
            code.send_keys(authcode)
            submit.click()

        #https://www.amazon.com/ref=gw_sgn_ib/853-0204854-22247543
        if "/ref=gw_sgn_ib/" in b.current_url:
            echo.out("Success, you are now signed in")
            b.save()


@arg('name', nargs=1, help="the name of the wishlist, amazon.com/gp/registry/wishlist/NAME")
@arg('--start-page', dest="start_page", type=int, default=1, help="The Wishlist page you want to start on")
@arg('--stop-page', dest="stop_page", type=int, default=0, help="The Wishlist page you want to stop on")
@arg('--debug', dest="debug", action=LoggingAction, help="Turn debugging on")
def main_dump(name, start_page, stop_page, **kwargs):
    """This is really here just to test that I can parse a wishlist completely and
    to demonstrate (by looking at the code) how to iterate through a list"""
    name = name[0]
    #pout.v(name, start_page, stop_page, kwargs)
    #pout.x()

    pages = set()
    current_url = ""
    w = Wishlist()
    for i, item in enumerate(w.get(name, start_page, stop_page), 1):
        new_current_url = w.current_url
        if new_current_url != current_url:
            current_url = new_current_url
            echo.h3(current_url)

        try:
            item_json = item.jsonable()
            echo.out("{}. {} is ${:.2f}", i, item_json["title"], item_json["price"])
            echo.indent(item_json["url"])

        except RobotError:
            raise

        except ParseError as e:
            echo.err("{}. Failed!", i)
            echo.err(e.body)
            echo.exception(e)

        except KeyboardInterrupt:
            break

        except Exception as e:
            echo.err("{}. Failed!", i)
            echo.exception(e)

        finally:
            pages.add(w.current_page)

    echo.out(
        "Done with wishlist, {} total pages parsed (from {} to {})",
        len(pages),
        start_page,
        stop_page
    )


if __name__ == "__main__":
    console()




"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.account import Account
from biweeklybudget.models.budget_model import Budget


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestTransStandingBudgetBalanceUpdate(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        """initial state verification"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 3
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert float(standing.current_balance) == 9482.29
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_1_add_trans_periodic_budget(self, testdb):
        """add a transaction against a periodic budget"""
        testdb.add(Transaction(
            actual_amount=222.22,
            budgeted_amount=123.45,
            description='T4',
            notes='notesT4',
            account=testdb.query(Account).get(1),
            budget=testdb.query(Budget).get(2)
        ))
        testdb.flush()
        testdb.commit()

    def test_2_verify_db(self, testdb):
        """verify no budget objects changed"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 4
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert float(standing.current_balance) == 9482.29
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_3_add_trans_standing_budget(self, testdb):
        """add a transaction against a standing budget"""
        testdb.add(Transaction(
            actual_amount=222.22,
            budgeted_amount=123.45,
            description='T5',
            notes='notesT5',
            account=testdb.query(Account).get(1),
            budget=testdb.query(Budget).get(5)
        ))
        testdb.flush()
        testdb.commit()

    def test_4_verify_db(self, testdb):
        """verify the standing budget balance was updated"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 5
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert float(standing.current_balance) == 9260.07
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_5_edit_trans_standing_budget(self, testdb):
        """edit a transaction against a standing budget"""
        t = testdb.query(Transaction).get(5)
        t.actual_amount = 111.11
        testdb.add(t)
        testdb.flush()
        testdb.commit()

    def test_6_verify_db(self, testdb):
        """verify the standing budget balance was updated"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 5
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert float(standing.current_balance) == 9371.18
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None

    def test_7_edit_trans_standing_budget(self, testdb):
        """edit a transaction against a standing budget"""
        t = testdb.query(Transaction).get(5)
        t.actual_amount = -111.11
        testdb.add(t)
        testdb.flush()
        testdb.commit()

    def test_8_verify_db(self, testdb):
        """verify the standing budget balance was updated"""
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 5
        standing = testdb.query(Budget).get(5)
        assert standing.is_periodic is False
        assert standing.name == 'Standing2'
        assert float(standing.current_balance) == 9593.40
        periodic = testdb.query(Budget).get(2)
        assert periodic.is_periodic is True
        assert periodic.name == 'Periodic2'
        assert periodic.current_balance is None
